#!/usr/bin/perl

# CLUSTERSIZE.PL -- a program to examine a FAT or NTFS boot sector
# and read the cluster size.
# (Written in Perl 4.036, and compatible with Perl 5.)
#
#
# This program is UNCOPYRIGHTED.  It was written by:
#
#    Dave Burton
#    Burton Systems Software
#    Tel: +1 919-481-0149
#    Email: dave at burtonsys.com
#


#  =>keyflag<=-- "%v, %d"
$ver_ordinal_and_date = "11, 05-Oct-09";
# the TLIB Version Control version number and last-modified date


$| = 1;   # predefined variable. If <> 0 then each print to the console
          # will immediatly be displayed, instead of buffered.


# Figure out what version of Perl we're running:
$perlver = "3 or earlier";
$perlvernum = 3;
if ($] =~ /\$\$Revision\:\s*([0-9.]+)\s/) {
   $perlver = $1;  # probably 4.something
   $perlvernum = int( 0 + $perlver );
} elsif ($] =~ /([0-9][0-9.]*)/) {
   $perlver = $1;  # probably 5.something or 6.something
   $perlvernum = int( 0 + $perlver );
}

# Figure out what OS it is running on:
if ($perlvernum <= 4) {
   # There is no $^O variable defined in Perl 4, but some versions of Perl 4
   # include OS information in the $] version string:
   if ($] =~ /for (NT|MSWin32)/) {
      $os = 'MSWin32';  # jperl 4.036 for Win32
   } else {
      $os = 'dos';  # just guessing that it is MS-DOS/PC-DOS/etc.
   }
} else {
   $os = $^O;
   if ('' eq $os) {
      # Perl prior to version 5.002 didn't support $^O
      if ($] =~ /for (NT|MSWin32)/) {
         $os = 'MSWin32';  # just a guess, in case it might have OS info in the $] variable, like Perl 4.036 did
      } elsif ('5.000' eq $]) {
         $os = 'dos';  # probably the unreleased DOS-extended Perl 5.000M5c that I still use in OS/2 DOS boxes
      } else {
         $os = '';  # I give up -- this means "unknown and unguessable"
      }
   }
   # possible values for $os are these standard $^O values:
   #   'aix', 'amigaos', 'beos', 'bsdos', 'cygwin', 'darwin', 'dec_osf', 'dgux',
   #   'dos', 'dynixptx', 'freebsd', 'hpux', 'irix', 'linux', 'machten', 'MacOS',
   #   'mpeix', 'MSWin32', 'NetWare', 'next', 'openbsd', 'openstep', 'os2',
   #   'os390', 'os400', 'posix-bc', 'rhapsody', 'riscos', 'sco_sv', 'solaris',
   #   'sunos', 'svr4', 'unicos',  'unicosmk', 'vmesa', 'VMS', 'VOS' (or 'vos'?)
   # and a null string that means "unknown":  '',
   # and probably others.
   # See:
   # http://www.google.com/search?q=%22This+is+Info+file+perl.info%2C+produced+by+Makeinfo%22
   # Get the latest/highest version, and search the document for mentions of $^O
   # e.g.,
   # http://search.cpan.org/src/KRISHPL/perl-5.6-info/info/perl.info-33
}



sub dump1sector {
   local( $sectr, $title ) = @_;
   local( $i, $ofs );
   if (defined $title) {
      print $title;
   }
   $i=$ofs=0;
   while ($ofs < 512) {
      printf "%03X ", $ofs;
      for ($j=0; $j < 24; $j++) {
         printf " %02X", ord(substr($sectr,$ofs,1));
         $ofs++;
         if ($ofs >= 512) {
            last;
         }
      }
      print "\n";
   }
}


# Perl pack/unpack format representing the structure of the FAT32 or NTFS Boot Sector:
$BootSector_fmt =
  'a3'  . # 000h   3 bytes  JMP 0x80h
  'a8'  . # 003h   8 bytes  Oem ID - Name of the formatting OS
          ## BIOS Parameter Block (00Bh - 03Fh):
  'S'   . # 00Bh   2 bytes  Bytes per Sector on the physical medium - Normally 512 bytes
  'C'   . # 00Dh   1 bytes  Sectors per Cluster - 1, 2, 4, 8, 16, 32, 64 or 128 sectors
  'S'   . # 00Eh   2 bytes  Reserved sectors in front of the FAT(s) incl. the Boot sector
  'C'   . # 010h   1 bytes  Number of FAT copies - Normaly 2 for FAT, 0 for NTFS
  'a4'  . # 011h   4 bytes  Not used in FAT32 or NTFS
  'C'   . # 015h   1 bytes  Media Descriptor - The same as in FAT16, but FAT32 and NTFS are only allowed on harddrives, so the value is F8h
  'S'   . # 016h   2 bytes  Not used in FAT32 or NTFS
  'S'   . # 018h   2 bytes  Sectors per Track - The disc geometry used when formatting the partition
  'S'   . # 01Ah   2 bytes  Heads - The disc geometry used when formatting the partition
  'L'   . # 01Ch   4 bytes  FAT32: Number of sectors on the disk from start of the partition to beginning of the first FAT
          #                 NTFS: Number of hidden sectors
  'a478'. # 020h 478 bytes  data that is not the same for NTFS and FAT
  'S'   ; # 1FEh   2 bytes  signature, 0xAA55


# Unpack a boot sector using $BootSector_fmt as the template:
sub unpack_a_boot_sector {
  local( $onesector ) = @_;
  local( @results ) = unpack( $BootSector_fmt, $onesector );
  return @results;
}

$verbose = 0;  # '-v' sets it to 1, -vv or "-v -v" sets it to 2
while (($#ARGV >= 0) && ($ARGV[0] =~ /^-v/)) {
   foreach $ch (split(//,$ARGV[0])) {
      if ('v' eq $ch) {
         $verbose++;
      }
   }
   shift @ARGV;
}
# if ($verbose) {
#    print "verbosity=$verbose\n";
# }

$numargs = $#ARGV + 1;
if (($numargs < 1) || ($numargs > 2)) {
   die   "clustersize.pl    v.$ver_ordinal_and_date\n" .
         "Display the cluster size(s) for FAT and NTFS file system(s)\n" .
         "\n" .
         "Usage:\n" .
         "   perl clustersize.pl drive\n" .
         "\n" .
         "or:\n" .
         "   perl clustersize.pl drive sect\n" .
         "\n" .
         "or:\n" .
         "   perl clustersize.pl -v drive sect\n" .
         "\n" .
         "Where:\n" .
         "  'drive' is a device name or a file containing a disk image.\n" .
         "  'sect'  is the integer LBA sector number (0-based) of the boot sector;\n" .
         "          if 'sect' is omitted, a full partition table is displayed.\n" .
         "  '-v'    verbose output (may be repeated 2 or 3 times for even more verbosity)\n" .
         "\n" .
         "Note: 'sect' is the 'Start' sector number displayed by 'fdisk -lu ...' under Linux\n";
}


$drive = $ARGV[0];
if ($numargs > 1) {
   $lba_offset1 = $ARGV[1];
   if ($lba_offset1 !~ /^[0-9]+$/) {
      print "ERROR: 2nd parameter (\"$lba_offset1\") must be an integer.\n";
      exit 2;
   }
}


$sectorsPerCluster = 0;  # side-effect result of &analyze() and &read_and_analyze()


# three input parameters are:
# 1) the 512-byte boot sector
# 2) the LBA disk offset of the boot sector
# 3) the verbosity level, 0-2:
#      0 = silent
#      1 = just print the number of sectors per cluster
#      2 = verbose output, including file system signature, etc.
sub analyze {
   local( $fbuf, $lba_offset, $verbose, $ptype ) = @_;
   $sectorsPerCluster = 0;  # side-effect result
   local( $fileofs ) = (512 * $lba_offset);
   local( $suggest_vv ) = 0;
   if ($verbose >= 2) {
      printf "LBA %d = 0x0%X  -->  File offset of boot sector is: %3.0f",
             $lba_offset, $lba_offset, $fileofs;
      if ($lba_offset < 8388608) {
         printf " = 0x%08X", $fileofs;
      } # else it would just overflow and print 0xFFFFFFFF
      print "\n";
   }
   local( @bootsectorfields ) = &unpack_a_boot_sector( $fbuf );
   local( $b_jmp80, $b_junk1, $b_junk2, $b_junk3, $b_junk4, $b_junk5, $b_junk6, $b_junk7, $b_junk8, $b_junk9, $b_AA55h );
   (  $b_jmp80          , # 000h   3 bytes  JMP 0x80h
      $OemId            , # 003h   8 bytes  Oem ID - Name of the formatting OS
                          ## BPB: BIOS Parameter Block (00Bh - 03Fh):
      $bytesPerSector   , # 00Bh   2 bytes  Bytes per Sector on the physical medium - Normally 512 bytes
      $sectorsPerCluster, # 00Dh   1 bytes  Sectors per Cluster - 1, 2, 4, 8, 16, 32, 64 or 128 sectors
      $b_junk1, $b_junk2, $b_junk3, $b_junk4, $b_junk5, $b_junk6, $b_junk7, $b_junk8, $b_junk9,
      $b_AA55h
   ) = @bootsectorfields;


   if ($verbose >= 2) {
      printf "000h  3 bytes  Jmp               = %02X %02X %02X\n", ord(substr($b_jmp80,0,1)), ord(substr($b_jmp80,1,1)), ord(substr($b_jmp80,2,1));
      printf "003h  8 bytes  Oem_Id            = '%-8s' (Oem ID - Name of formatting OS)\n", $OemId;
      printf "00Bh  2 bytes  bytesPerSector    = %-6d     (Bytes per Sector - Normally 512)\n", $bytesPerSector;
      printf "00Dh  1 bytes  sectorsPerCluster = %-4d       (Sectors per Cluster)\n", $sectorsPerCluster;
      printf "1FEh  2 bytes  AA55h             = 0x%04X     (signature, should be 0xAA55)\n", $b_AA55h;
   }
   # else $verbose==0 so &analyze() is silent.

   # for bootable partition types, do sanity checks on a few fields
   if ($verbose >= 4) {
      print "dbg: ptype='$ptype'\n";
   }
   if (",5,f,82,83,fd," !~ ",$ptype,") {
      if ($verbose >= 4) {
         print "  '$ptype' is not in {,5,f,82,83,fd,} so try to find sectors/cluster...\n";
      }

      # Here's where we (finally) display the number of sectors per cluster:
      if ($verbose >= 1) {
        printf " %d sectors per cluster\n", $sectorsPerCluster;
      }

      # for bootable partition types, do sanity checks on a few fields:
      if ($b_AA55h != 0x0AA55) {
         printf "ERROR: boot sector signature is 0x%04X, but should be 0xAA55.\n", $b_AA55h;
         $suggest_vv++;
      }
      if (512 != $bytesPerSector) {
         printf "ERROR: bytes per sector is %d, but should be 512.\n", $bytesPerSector;
         $suggest_vv++;
      }
      # check for illegal sectors-per-cluster value:
      $badspc = 1;
      foreach $spc (1,2,4,8,16,32,64,128) {
         if ($sectorsPerCluster == $spc) {
            $badspc = 0;
         }
      }
      if ($badspc) {
         printf "ERROR: sectors per cluster is %d, but should be in {1,2,4,8,16,32,64,128}.\n",
               $sectorsPerCluster;
         $suggest_vv++;
      }
      if ($suggest_vv > 1) {
         if ($b_AA55h != 0x0AA55) {
            print "*** THIS DOES NOT APPEAR TO BE A BOOT SECTOR ***\n";
         } elsif (0==$lba_offset) {
            print "*** THIS LOOKS MORE LIKE AN MBR THAN A FILE SYSTEM BOOT SECTOR ***\n";
         }
      }
   }

   if (($verbose >= 3) || (($verbose >= 2) && $suggest_vv)) {
      &dump1sector( $fbuf, "Boot sector is:\n" );
   } elsif ($suggest_vv) {
      print "(To display the entire boot sector, specify the '-vvv' parameter.)\n";
   }

   return $sectorsPerCluster;
}


$diskimage_is_open = 0;

# open the drive or image file:
sub open_diskimage {
   local( $drive ) = @_;
   if (!open( DISKIMAGE, "<$drive" )) {
      print "ERROR: Could not open '$drive', $!\n";
      return 2;  # error
   }
   binmode DISKIMAGE;
   $diskimage_is_open = 1;
   return 0;  # success
}


sub read_and_analyze {
   local( $lba_offset, $ptype ) = @_;
   local( $fileofs ) = (512 * $lba_offset);
   local( $rc ) = 0;  # no error yet
   local( $fsectbuf );
   seek( DISKIMAGE, $fileofs, 0 );
   if (!sysread(DISKIMAGE, $fsectbuf, 512)) {
      print "ERROR: Could not read '$drive' at ofs=$fileofs, $!\n";
      $rc = 2;  # error
      $OemId = '????????';
      $bytesPerSector = $sectorsPerCluster = 0;
   } else {
      &analyze( $fsectbuf, $lba_offset, $verbose, $ptype );
   }
   return $rc;
}


# input is the output of 'fdisk -lu ...' which looks something like this:
#
# ---------------------------------------------------------------------
# You must set cylinders.
# You can do this from the extra functions menu.
#
# Disk drive.ima: 0 MB, 0 bytes
# 255 heads, 63 sectors/track, 0 cylinders, total 0 sectors
# Units = sectors of 1 * 512 = 512 bytes
#
#     Device Boot      Start         End      Blocks   Id  System
# drive.ima1              63       80324       40131   de  Dell Utility
# drive.ima2   *       80325    60050969    29985322+   7  HPFS/NTFS
# Partition 2 has different physical/logical endings:
#     phys=(1023, 254, 63) logical=(3737, 254, 63)
# ---------------------------------------------------------------------
#
sub process_all_partitions {
   local( @fdisk_out ) = @_;
   local( @words );
   local( $numsucceeded ) = 0;
   if (0==$verbose) {
      print "part#  Boot     Start         End      Blocks   Id  Oem-Id   Sec/Cluster  System\n";
   }
   foreach $line (@fdisk_out) {
      @words = split( /\s+/, $line );
      if ($#words >= 4) {
         $bootind = ' ';
         if ('*' eq $words[1]) {
            splice( @words, 1, 1 );  # delete the '*' boot indicator
            $bootind = '*';
         }
         if ($words[0] =~ /^([\w\/\.\$]*[a-zA-Z])([0-9]+)$/) {
            # Device looks reasonable, and ends in a digit, as it should
            $device_or_image_file_name = '';  # shut up a Perl warning
            $device_or_image_file_name = $1;  # either $drive or $drive.'p'
            $partition_number = $2;  # the digit(s) at the end
            if (($words[1] =~ /^\d+$/) && ($words[2] =~ /^\d+$/)) {
               # Start and End both appear to be numeric
               $start = 0 + $words[1];
               $end = 0 + $words[2];
               if ($#words >= 4) {
                  # even after boot indicator removed, there're still enough words to be reasonable
                  if ($words[3] =~ /^(\d+)(\+|)$/) {
                     # Blocks count seems reasonable (the possible '+' at the end means '.5')
                     $blocks = $1;
                     if ('+' eq $2) {
                        $blocks .= '.5';
                     } else {
                        $blocks .= '  ';
                     }
                     if ($words[4] =~ /^[0-9a-fA-F]{1,2}$/) {
                        # Id (partition type) is 1 or two hexadecimal digits, as it should be
                        $ptype = $words[4];
                        $system_name = '';
                        if ($#words >= 5) {
                           $system_name = join( ' ', splice(@words,5) );
                        }
                        $OemId = '????????';  # i.e., unknown
                        $sectorsPerCluster = 0;  # i.e., unknown
                        if ($verbose) {
                           print "-------------------------------------------------------------------------------\n";
                        }
                        if ($verbose >= 2) {
                           print "part#  Boot     Start         End      Blocks    Id  System\n";
                           printf " %2d     %s%12d%12d%14s  %2s  %s\n",
                             $partition_number, $bootind, $start, $end, $blocks, $ptype, $system_name;
                        }
                        if ($diskimage_is_open) {
                           if (0 == &read_and_analyze( $start, $ptype )) {
                              $numsucceeded++;
                           }
                        }
                        $tmpOemId = $OemId;  # side-effect result of &analyze()
                        $tmpOemId =~ s/[\s\x00]+$//g;  # remove trailing blanks and/or null bytes
                        if ('' eq $tmpOemId) {
                           $tmpOemId = '        ';
                        }
                        $tmpOemId =~ tr/\x20-\x7E/\?/c;  # change all unprintable characters to '?'

                        if ($verbose >= 4) {
                           print "dbg: OemID='$OemId' =";
                           foreach $ch (split(//,$OemId)) {
                              printf " 0x%02X", ord($ch);
                           }
                           print " -> '$tmpOemId'\n";
                        }

                        if ($verbose > 0) {
                           print "part#  Boot     Start         End      Blocks   Id  Oem-Id   Sec/Cluster  System\n";
                        }

                        printf " %2d     %s%12d%12d%14s %2s  %-8s    ",
                               $partition_number, $bootind, $start, $end, $blocks, $ptype, $tmpOemId;
                        if ($sectorsPerCluster > 0) {
                           printf "%3d", $sectorsPerCluster;
                        } else {
                           print  "  -";
                        }
                        print "       $system_name\n";
                     }
                  }
               }
            }
         }
      }
   }
   return $numsucceeded;
}


$rc = 0;
if (2 == $numargs) {
   $rc = &open_diskimage( $drive );
   if (!$rc) {
      $rc = &read_and_analyze( $lba_offset1, '4' );  # ptype=0x04 to force analysis (pass as 1 or 2-byte string)
      if (!$verbose) {
         print "$sectorsPerCluster\n";
      }
   }
} else {
   $quoted_drive = $drive;
   if ($drive =~ /\s/) {
      # in case image file name contains spaces
      $quoted_drive = "\"$drive\"";
   }
   if (! -e $drive) {
      $rc = 2;
      print "ERROR: $quoted_drive does not exist.\n";
      # Unfortunately, this doesn't detect the case of a nonexistant device,
      # such as /dev/hdb when there's only one drive and it's /dev/hda
   } else {
      if ($verbose >= 4) {
         print "dbg: '-e $quoted_drive' = " . (-e $drive) . "\n";
         print "dbg: '-b $quoted_drive' = " . (-b $drive) . "\n";
         print "dbg: '-r $quoted_drive' = " . (-r $drive) . "\n";
         print "dbg: '-R $quoted_drive' = " . (-R $drive) . "\n";
         print "dbg: '-s $quoted_drive' = " . (-s $drive) . "\n";
         print "dbg: '-z $quoted_drive' = " . (-z $drive) . "\n";
         # Drat, these all seem to print the same things (1,1,1,1,0,1) regardless
         # of whether or not the device actually exists!
      }
      if (!open( FDISK, "fdisk -lu $quoted_drive 2>/dev/null 1|" )) {
         # this never seems to happen
         $rc = 2;
         print "ERROR: could not run \"fdisk -lu $quoted_drive\".  $!\n";
      } else {
         @fdisk_lu_output = <FDISK>;
         close FDISK;
         if ($#fdisk_lu_output < 0) {
            # This is what seems to happen under Windows 2000, and also what
            # seems to happen under Linux if a missing device is specified.
            $rc = 2;
            print "ERROR: \"fdisk -lu $quoted_drive\" failed.  $!\n";
         }
      }
      if ($rc >= 2) {
         if ($os !~ /^linux$/) {
            # print "dbg: You are running $os, not Linux.\n";
            if ($os =~ /^(MSWin32|dos)$/) {
               print "That's because 'fdisk -lu' does not work under DOS or Windows.  Try using\n";
            } else {
               print "Perhaps 'fdisk -lu' does not work under $os, of maybe $quoted_drive doesn't\n".
                     "exist.  Try using ";
            }
         } else {
            print "Perhaps $quoted_drive does not exist, or perhaps 'fdisk -lu' is broken.  Try\n" .
                  "using ";
         }
         print "the sect parameter to specify a particular boot sector (maybe 63).\n";
      }
      if (!$rc) {
         foreach $line (@fdisk_lu_output) {
            $line =~ s/\s*$//;  # remove newlines & trailing whitespace
         }
         if ($verbose) {
            print "fdisk -lu: " . join("\nfdisk -lu: ",@fdisk_lu_output) . "\n";
         }
         $rc = &open_diskimage( $drive );
         if (0 == $rc) {
            $numsucceeded = &process_all_partitions( @fdisk_lu_output );
            if ((!$rc) && (0 == $numsucceeded)){
               $rc = 1;
            }
         }
      }
   }
}

if ($verbose >= 2) {
   print "exit w/ errorlevel $rc\n";
}
exit( $rc );

__END__

